import numpy as np

STATUS_FAILURE = 1
STATUS_SUCCESS = 2


class Trials(object):
    """
    A class for storing trials data.
    """

    def __init__(self):
        self.trials = []
        self.lowest_loss = np.inf

        self.best_params = None
        self.last_params = None

        # instance variables for optimizers
        self._train_x = None
        self._train_y = None

    def __len__(self):
        return len(self.trials)

    def get_train_data(self):
        """
        A method to get train data.
        The train data is arranged in alphabetical order of
        the name of parameters.
        """

        return self._train_x, self._train_y

    def add(self, result, params, exp_info=None):
        """
        A method to add each trial result
        """

        loss = result.get('loss', np.inf)

        t = {
            'result': result,
            'loss': loss,
            'params': params,
            'exp_key': exp_info['exp_key'],
            'eval_count': exp_info['eval_count'],
        }

        if result['status'] == STATUS_SUCCESS:
            t['status'] = STATUS_SUCCESS

            # Because the iteration order is not determinstic,
            # sort params dictionary by alphabetical order.
            train_x_row = []
            for key in sorted(params):
                train_x_row.append(params[key])

            if self._train_x is None:
                self._train_x = np.array([train_x_row])
            else:
                self._train_x = np.vstack((self._train_x, train_x_row))

            if self._train_y is None:
                self._train_y = np.array([loss])
            else:
                self._train_y = np.hstack((self._train_y, loss))
        else:
            t['status'] = STATUS_FAILURE

        # Update last_params
        last_params = []
        for k in sorted(params):
            last_params.append(params[k])
        self.last_params = np.array(last_params)

        # Set lowest loss to calculate expected improvements.
        if loss < self.lowest_loss:
            self.lowest_loss = loss
            self.best_params = params

        self.trials.append(t)

        return





def calc_next_params(domain, trials):
    next_params = {}
    random_params = domain.random()
    for index, fieldname in enumerate(domain.fieldnames):
        next_params[fieldname] = random_params[index]
    return [next_params]



from polaris import Polaris, Trials, Bounds
from polaris.examples.utils import pseudo_train


if __name__ == '__main__':
    bounds = [
        Bounds('lr', 0.001, 0.01),
        Bounds('weight_decay', 0.0002, 0.04),
    ]
    trials = Trials()
    polaris = Polaris(
            pseudo_train, bounds, 'bayesian_opt',
            trials, max_evals=100, debug=True, exp_key='this_is_test')
    best_params = polaris.run(use_mpi=True)
    print('best params: ', best_params)



import chainer
import chainer.functions as F
import chainer.links as L


class LeNet5(chainer.Chain):

    def __init__(self, n_class=10):
        super(LeNet5, self).__init__()
        with self.init_scope():
            self.conv1 = L.Convolution2D(3, 32, 5, stride=1, pad=2)
            self.conv2 = L.Convolution2D(32, 32, 5, stride=1, pad=2)
            self.conv3 = L.Convolution2D(32, 64, 5, stride=1, pad=2)
            self.fc4 = L.Linear(None, 4096)
            self.fc5 = L.Linear(4096, n_class)

    def __call__(self, x):
        h = F.max_pooling_2d(F.relu(self.conv1(x)), 3, stride=2)
        h = F.max_pooling_2d(F.relu(self.conv2(h)), 3, stride=2)
        h = F.relu(self.conv3(h))
        h = F.spatial_pyramid_pooling_2d(h, 3, F.MaxPooling2D)
        h = F.relu(self.fc4(h))
        h = self.fc5(h)
        return h



import argparse

from polaris import Polaris, Trials, Bounds

BOUNDS = {
    'momentum_sgd': [
            Bounds('weight_decay', 0, 0.001),
            Bounds('lr', 0.005, 0.02),
            Bounds('lr_decay_rate', 0.3, 0.7),
            Bounds('lr_decay_epoch', 30, 80, 10),
            Bounds('momentum', 0.6, 0.98),
        ],
    'adam': [
            Bounds('weight_decay', 0, 0.001),
            Bounds('alpha', 0, 0.005),
            Bounds('beta1', 0.6, 0.95),
            Bounds('beta2', 0.6, 0.9999),
            Bounds('eta', 0.9, 1.0),
            Bounds('eps', 1e-9, 1e-10),
        ],
    }

if __name__ == '__main__':
    import optimizer

    parser = argparse.ArgumentParser(description='Chainer CIFAR example:')

    # Set model
    parser.add_argument('--arch', '-a', type=str, default='resnet50')

    # Set batchsize
    parser.add_argument('--batchsize', '-B', type=int, default=128)

    # Set running environment
    parser.add_argument('--communicator', type=str, default='pure_nccl')
    parser.add_argument('--epoch', '-e', type=int, default=250)
    parser.add_argument('--out', '-o', default='result')
    parser.add_argument('--seed', type=int, default=0)
    parser.add_argument('--optimizer_name', type=str)

    # Data augmentation settings
    parser.add_argument('--random_angle', type=float, default=15.0)
    parser.add_argument('--pca_sigma', type=float, default=25.5)
    parser.add_argument('--expand_ratio', type=float, default=1.2)
    parser.add_argument('--crop_size', type=int, nargs='*', default=[28, 28])

    args = parser.parse_args()
    bounds = BOUNDS[args.optimizer_name]

    job_name = f'cifar10_mn_example'
    trials = Trials()

    polaris = Polaris(
            optimizer.run, bounds, 'bayesian_opt',
            trials, max_evals=5, exp_key=job_name, args=(args,))

    best_params = polaris.run_parallel()
    print(best_params)



import os
from urllib.parse import urlparse

RABBITMQ_URL = os.getenv('RABBITMQ_URL', 'amqp://localhost//')
url = urlparse(RABBITMQ_URL)

RABBITMQ_HOST = url.hostname
RABBITMQ_USERNAME = url.username
RABBITMQ_PASSWORD = url.password
RABBITMQ_PORT = 5672
RABBITMQ_VIRTUAL_HOST = url.path[1:]



import numpy as np
import scipy as sp
import statsmodels.nonparametric.kernel_density as kde


DEFAULT_GAMMA = 0.25
DEFAULT_BW = None
DEFAULT_N_MIN = 8
DEFAULT_BW_WEIGHT = 3.
DEFAULT_SAMPLING_NUM = 64


def objective_function(x, l, g):
    return -l.pdf(x) / g.pdf(x)


def trunc_range(a, b, m, s):
    # ref. https://docs.scipy.org/doc/scipy/reference/
    #      generated/scipy.stats.truncnorm.html
    return (a - m) / s, (b - m) / s


def minimize(
    fun,
    x,
    sampling_num,
    bw,
    bounds,
    args=None
):
    best = np.inf
    ret = None
    for p in x[np.random.choice(x.shape[0], sampling_num)]:
        sample = []
        for mean, sigma, (a, b) in zip(p, bw, bounds):
            a, b = trunc_range(a, b, mean, sigma)
            sample.append(sp.stats.truncnorm.rvs(a, b, loc=mean, scale=sigma))
            # ref. https://en.wikipedia.org/wiki/Truncated_normal_distribution
        val = fun(sample, *args)
        if not np.isfinite(val):
            continue
        if val < best:
            best = val
            ret = sample
    return ret


def calc_next_params(domain, trials):

    gamma = DEFAULT_GAMMA
    bw = DEFAULT_BW
    n_min = DEFAULT_N_MIN
    bw_weight = DEFAULT_BW_WEIGHT
    sampling_num = DEFAULT_SAMPLING_NUM

    next_params = {}
    if len(trials) <= n_min+2:
        random_result = domain.random()
        for index, fieldname in enumerate(domain.fieldnames):
            next_params[fieldname] = random_result[index]
        return next_params

    train_x, train_y = trials.get_train_data()
    idx = np.argsort(train_y)
    n = len(trials)
    l_len = max(n_min, int(n*gamma))
    g_len = max(n_min, n-l_len)
    x_l = train_x[idx[:l_len], :]
    x_g = train_x[idx[-g_len:], :]

    # I want to get the types of params from domain
    # v_types = "ccccuuuuuooooo"
    # ref. http://www.statsmodels.org/dev/generated/
    #      statsmodels.nonparametric.kernel_density.KDEMultivariate.html

    v_types = 'c' * domain.n_params

    l_est = kde.KDEMultivariate(
        x_l,
        v_types,
        bw=bw
    )

    g_est = kde.KDEMultivariate(
        x_g,
        v_types,
        bw=bw
    )

    wide_bw = l_est.bw * bw_weight
    for w in np.nditer(wide_bw, op_flags=['readwrite']):
        w[...] = max(w, 1e-3*bw_weight)
    bounds = domain.bounds
    minimize_result = minimize(
        fun=objective_function,
        x=x_l,
        sampling_num=sampling_num,
        bw=wide_bw,
        bounds=bounds,
        args=(l_est, g_est)
    )

    for index, fieldname in enumerate(domain.fieldnames):
        next_params[fieldname] = minimize_result[index]

    return next_params
